/*******************************************************************************
Copyright (c) 2011, Luis Lourenco
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*******************************************************************************/
/* 
 * Second Algorithm Test - Warp Serialization and Memory Access
**/

#include <stdio.h>
#include <assert.h>
#include <cuda.h>
#include <cutil.h>

texture<float, 1, cudaReadModeElementType> texRef;

__global__ void kernel(float *out, float *in, int N)
{
  int idx = blockIdx.x*blockDim.x + threadIdx.x;
  float2 val;
  if(idx>0){
    val.x = in[idx-1];
  }
  else{
    val.x = in[idx];
  }
  if(idx<N-1){
    val.y = in[idx+1];
  }
  else{
    val.y = in[idx];
  }
    
  if (idx<N) out[idx] = sqrt((2.0*val.x)+(0.5*val.y));
}

__global__ void fastkernel(float *out, int N)
{
  int idx = blockIdx.x*blockDim.x + threadIdx.x;
  float2 val;
  if(idx>0){
    val.x = tex1Dfetch(texRef, idx-1);
  }
  else{
    val.x = tex1Dfetch(texRef, idx);
  }
  if(idx<N-1){
    val.y = tex1Dfetch(texRef, idx+1);
  }
  else{
    val.y = tex1Dfetch(texRef, idx);
  }

  if (idx<N) out[idx] = sqrt((2.0*val.x)+(0.5*val.y));

}

__global__ void fastestkernel(float *out, int N)
{
  int idx = blockIdx.x*blockDim.x + threadIdx.x;
  if (idx<N){
    out[idx] = sqrt(2.0*tex1Dfetch(texRef, idx-(idx>0))+(0.5*tex1Dfetch(texRef, idx+(idx<N-1))));
  }
}


int main(void)
{
  cudaSetDevice(1);
  //cudaThreadSetCacheConfig(cudaFuncCachePreferL1);
  float *orig_h, *orig_d;
  float *cpu_h, *gpu_h, *opt_h;           // pointers to host memory
  float *gpu_d, *opt_d;                 // pointer to device memory
  int i, N = 10000000;
  size_t size = N*sizeof(float);
  // allocate arrays on host
  orig_h = (float *)malloc(size);
  cpu_h = (float *)malloc(size);
  gpu_h = (float *)malloc(size);
  opt_h = (float *)malloc(size);
  // allocate array on device 
  cudaMalloc((void **) &orig_d, size);
  cudaMalloc((void **) &gpu_d, size);
  cudaMalloc((void **) &opt_d, size);
  // initialization of host data
  srand ( time(NULL) );
  for (i=0; i<N; i++) orig_h[i] = (float)(rand() % 1000);
  // copy data from host to device
  cudaMemcpy(orig_d, orig_h, sizeof(float)*N, cudaMemcpyHostToDevice);
  // do calculation on host
  // Part 1 of 2. Compute execution configuration
  int blockSize = 256;
  int nBlocks = N/blockSize + (N%blockSize == 0?0:1);
  ///Warmup
  unsigned int WarmupTimer = 0;
  cutCreateTimer( &WarmupTimer );
  cutStartTimer( WarmupTimer );
  float *rub;
  cudaMalloc( (void**)&rub, size * sizeof(float) );
  cudaFree( rub );
  CUT_CHECK_ERROR("Warmup failed");
  kernel <<< nBlocks, blockSize >>> (gpu_d, orig_d, N);
  cudaThreadSynchronize();
  cutStopTimer( WarmupTimer );
  printf("Warmup time = %f ms\n",cutGetTimerValue( WarmupTimer ));
  unsigned int CannyTimer = 0;
  cutCreateTimer( &CannyTimer );
  // Create timer for timing CUDA calculation
  unsigned int cudaTimer = 0;
  unsigned int cudaTimer2 = 0;
  unsigned int cudaTimer3 = 0;
  cutCreateTimer( &cudaTimer );
  cutCreateTimer( &cudaTimer2 );
  cutCreateTimer( &cudaTimer3 );
  // Part 2 of 2. Call incrementArrayOnDevice kernel 
  cutStartTimer( cudaTimer );  // Start timer
  kernel <<< nBlocks, blockSize >>> (gpu_d, orig_d, N);
  cudaThreadSynchronize();  // Wait for square_array to finish on CUDA
  cutStopTimer( cudaTimer );  // Stop timer
  // Part 3 of 3. Call incrementArrayOnDeviceOptimized kernel 
  cudaBindTexture (NULL ,texRef, orig_d);
  texRef.normalized = false;
  texRef.filterMode = cudaFilterModePoint;
  CUT_CHECK_ERROR("Texture bind failed");
  cutStartTimer( cudaTimer2 );  // Start timer
  fastkernel <<< nBlocks, blockSize >>> (opt_d, N);
  cudaThreadSynchronize();  // Wait for square_array to finish on CUDA
  cutStopTimer( cudaTimer2 );  // Stop timer
  cudaUnbindTexture(texRef);
  CUT_CHECK_ERROR("Memory image free failed");
  // Part 3 of 3. Call incrementArrayOnDeviceOptimized kernel 
  cudaBindTexture (NULL ,texRef, orig_d);
  texRef.normalized = false;
  texRef.filterMode = cudaFilterModePoint;
  CUT_CHECK_ERROR("Texture bind failed");
  cutStartTimer( cudaTimer3 );  // Start timer
  fastestkernel <<< nBlocks, blockSize >>> (opt_d, N);
  cudaThreadSynchronize();  // Wait for square_array to finish on CUDA
  cutStopTimer( cudaTimer3 );  // Stop timer
  cudaUnbindTexture(texRef);
  CUT_CHECK_ERROR("Memory image free failed");
  // Retrieve result from device and store in b_h
  cudaMemcpy(gpu_h, gpu_d, sizeof(float)*N, cudaMemcpyDeviceToHost);
  cudaMemcpy(opt_h, opt_d, sizeof(float)*N, cudaMemcpyDeviceToHost);
  // check results
  for (i=0; i<N; i++) assert(gpu_h[i] == opt_h[i]);
//  for (i=0; i<N; i++) printf("%f == %f\n",gpu_h[i], opt_h[i]);
  printf("Naive: %f\n",cutGetTimerValue( cudaTimer ));
  printf("Opt1: %f\n",cutGetTimerValue( cudaTimer2 ));
  printf("Opt2: %f\n",cutGetTimerValue( cudaTimer3 ));
  // cleanup
  free(orig_h);
  free(cpu_h);
  free(gpu_h);
  free(opt_h);
  cudaFree(orig_d); 
  cudaFree(gpu_d); 
  cudaFree(opt_d); 
}
